# Copyright 2012-2013 NEC Corporation.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import helpers as log_helpers
from oslo_log import log as logging
from oslo_utils import importutils

from neutron.agent import securitygroups_rpc as sg_rpc
from neutron.api import extensions as neutron_extensions
from neutron.api.rpc.agentnotifiers import dhcp_rpc_agent_api
from neutron.api.rpc.handlers import dhcp_rpc
from neutron.api.rpc.handlers import l3_rpc
from neutron.api.rpc.handlers import metadata_rpc
from neutron.api.rpc.handlers import securitygroups_rpc
from neutron.common import constants as const
from neutron.common import exceptions as n_exc
from neutron.common import rpc as n_rpc
from neutron.common import topics
from neutron.db import agents_db
from neutron.db import agentschedulers_db
from neutron.db import allowedaddresspairs_db as addr_pair_db
from neutron.db import db_base_plugin_v2
from neutron.db import external_net_db
from neutron.db import portbindings_base
from neutron.db import quota_db  # noqa
from neutron.extensions import allowedaddresspairs as addr_pair
from neutron.plugins.common import constants as svc_constants

from networking_nec.plugins.openflow import config as nec_config
from networking_nec.plugins.openflow import extensions
from networking_nec.plugins.openflow import l2manager
from networking_nec.plugins.openflow import ofc_manager
from networking_nec.plugins.openflow import packet_filter
from networking_nec.plugins.openflow import portbindings as bindings
from networking_nec.plugins.openflow import router as router_plugin
from networking_nec.plugins.openflow import rpc

LOG = logging.getLogger(__name__)


class NECPluginV2(db_base_plugin_v2.NeutronDbPluginV2,
                  external_net_db.External_net_db_mixin,
                  router_plugin.RouterMixin,
                  rpc.SecurityGroupServerRpcMixin,
                  agentschedulers_db.DhcpAgentSchedulerDbMixin,
                  router_plugin.L3AgentSchedulerDbMixin,
                  packet_filter.PacketFilterMixin,
                  bindings.PortBindingMixin,
                  addr_pair_db.AllowedAddressPairsMixin):

    _supported_extension_aliases = ["agent",
                                    "allowed-address-pairs",
                                    "binding",
                                    "dhcp_agent_scheduler",
                                    "external-net",
                                    "ext-gw-mode",
                                    "extraroute",
                                    "l3_agent_scheduler",
                                    "packet-filter",
                                    "quotas",
                                    "router",
                                    "router_provider",
                                    "security-group",
                                    ]

    @property
    def supported_extension_aliases(self):
        if not hasattr(self, '_aliases'):
            aliases = self._supported_extension_aliases[:]
            self.setup_extension_aliases(aliases)
            self._aliases = aliases
        return self._aliases

    def setup_extension_aliases(self, aliases):
        sg_rpc.disable_security_group_extension_by_config(aliases)
        self.remove_packet_filter_extension_if_disabled(aliases)

    def __init__(self):
        nec_config.register_plugin_opts()
        super(NECPluginV2, self).__init__()
        self.ofc = ofc_manager.OFCManager(self.safe_reference)
        self.l2mgr = l2manager.L2Manager(self.safe_reference)
        self.base_binding_dict = self._get_base_binding_dict()
        portbindings_base.register_port_dict_function()

        neutron_extensions.append_api_extensions_path(extensions.__path__)

        self.setup_rpc()
        self.l3_rpc_notifier = router_plugin.L3AgentNotifyAPI()

        self.network_scheduler = importutils.import_object(
            cfg.CONF.network_scheduler_driver
        )
        self.router_scheduler = importutils.import_object(
            cfg.CONF.router_scheduler_driver
        )

        router_plugin.load_driver(self.safe_reference, self.ofc)
        self.start_periodic_dhcp_agent_status_check()

    def setup_rpc(self):
        self.service_topics = {svc_constants.CORE: topics.PLUGIN,
                               svc_constants.L3_ROUTER_NAT: topics.L3PLUGIN}
        self.conn = n_rpc.create_connection(new=True)
        self.notifier = rpc.NECPluginV2AgentNotifierApi(topics.AGENT)
        self.agent_notifiers[const.AGENT_TYPE_DHCP] = (
            dhcp_rpc_agent_api.DhcpAgentNotifyAPI()
        )
        self.agent_notifiers[const.AGENT_TYPE_L3] = (
            router_plugin.L3AgentNotifyAPI()
        )

        # NOTE: callback_sg is referred to from the sg unit test.
        self.callback_sg = securitygroups_rpc.SecurityGroupServerRpcCallback()
        self.endpoints = [
            rpc.NECPluginV2RPCCallbacks(self.safe_reference),
            dhcp_rpc.DhcpRpcCallback(),
            l3_rpc.L3RpcCallback(),
            self.callback_sg,
            agents_db.AgentExtRpcCallback(),
            metadata_rpc.MetadataRpcCallback()]
        for svc_topic in self.service_topics.values():
            self.conn.create_consumer(svc_topic, self.endpoints, fanout=False)
        # Consume from all consumers in threads
        self.conn.consume_in_threads()

    @log_helpers.log_method_call
    def create_network(self, context, network):
        """Create a new network entry on DB, and create it on OFC."""
        tenant_id = self._get_tenant_id_for_create(context, network['network'])
        # set up default security groups
        self._ensure_default_security_group(context, tenant_id)
        network['network']['status'] = self.l2mgr.get_initial_net_status(
            network)
        with context.session.begin(subtransactions=True):
            new_net = super(NECPluginV2, self).create_network(context,
                                                              network)
            self._process_l3_create(context, new_net, network['network'])

        self.l2mgr.create_network(context, new_net)

        return new_net

    @log_helpers.log_method_call
    def update_network(self, context, id, network):
        """Update network and handle resources associated with the network.

        Update network entry on DB. If 'admin_state_up' was changed, activate
        or deactivate ports and packetfilters associated with the network.
        """

        session = context.session
        with session.begin(subtransactions=True):
            old_net = super(NECPluginV2, self).get_network(context, id)
            new_net = super(NECPluginV2, self).update_network(context, id,
                                                              network)
            self._process_l3_update(context, new_net, network['network'])

        self.l2mgr.update_network(context, id, old_net, new_net)

        return new_net

    @log_helpers.log_method_call
    def delete_network(self, context, id):
        """Delete network and packet_filters associated with the network.

        Delete network entry from DB and OFC. Then delete packet_filters
        associated with the network. If the network is the last resource
        of the tenant, delete unnessary ofc_tenant.
        """
        ports = self.get_ports(context, filters={'network_id': [id]})

        # check if there are any tenant owned ports in-use;
        # consider ports owned by floating ips as auto_delete as if there are
        # no other tenant owned ports, those floating ips are disassociated
        # and will be auto deleted with self._process_l3_delete()
        only_auto_del = all(p['device_owner'] in
                            db_base_plugin_v2.AUTO_DELETE_PORT_OWNERS or
                            p['device_owner'] == const.DEVICE_OWNER_FLOATINGIP
                            for p in ports)
        if not only_auto_del:
            raise n_exc.NetworkInUse(net_id=id)

        self._process_l3_delete(context, id)

        self.l2mgr.delete_network(context, id, ports)

        super(NECPluginV2, self).delete_network(context, id)

    @log_helpers.log_method_call
    def create_port(self, context, port):
        """Create a new port entry on DB, then try to activate it."""

        port['port']['status'] = self.l2mgr.get_initial_port_status(port)
        port_data = port['port']
        with context.session.begin(subtransactions=True):
            self._ensure_default_security_group_on_port(context, port)
            sgids = self._get_security_groups_on_port(context, port)
            new_port = super(NECPluginV2, self).create_port(context, port)
            self._process_portbindings_create(context, port_data, new_port)
            self._process_port_create_security_group(
                context, new_port, sgids)
            new_port[addr_pair.ADDRESS_PAIRS] = (
                self._process_create_allowed_address_pairs(
                    context, new_port,
                    port_data.get(addr_pair.ADDRESS_PAIRS)))
        self.notify_security_groups_member_updated(context, new_port)
        return self.l2mgr.create_port(context, new_port)

    @log_helpers.log_method_call
    def update_port(self, context, id, port):
        """Update port, and handle packetfilters associated with the port.

        Update network entry on DB. If admin_state_up was changed, activate
        or deactivate the port and packetfilters associated with it.
        """
        need_port_update_notify = False
        with context.session.begin(subtransactions=True):
            old_port = super(NECPluginV2, self).get_port(context, id)
            new_port = super(NECPluginV2, self).update_port(context,
                                                            id, port)
            self._process_portbindings_update(context, port['port'], new_port)
            if addr_pair.ADDRESS_PAIRS in port['port']:
                need_port_update_notify |= (
                    self.update_address_pairs_on_port(context, id, port,
                                                      old_port,
                                                      new_port))
            need_port_update_notify |= self.update_security_group_on_port(
                context, id, port, old_port, new_port)

        need_port_update_notify |= self.is_security_group_member_updated(
            context, old_port, new_port)
        if need_port_update_notify:
            self.notifier.port_update(context, new_port)

        self.l2mgr.update_port(context, old_port, new_port)
        return new_port

    @log_helpers.log_method_call
    def delete_port(self, context, id, l3_port_check=True):
        """Delete port and packet_filters associated with the port."""
        # if needed, check to see if this is a port owned by
        # and l3-router.  If so, we should prevent deletion.
        if l3_port_check:
            self.prevent_l3_port_deletion(context, id)

        port = self.l2mgr.delete_port(context, id)

        with context.session.begin(subtransactions=True):
            router_ids = self.disassociate_floatingips(
                context, id, do_notify=False)
            super(NECPluginV2, self).delete_port(context, id)

        # now that we've left db transaction, we are safe to notify
        self.notify_routers_updated(context, router_ids)
        self.notify_security_groups_member_updated(context, port)



# Copyright 2012 NEC Corporation.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_utils import importutils

from networking_nec._i18n import _LI


LOG = logging.getLogger(__name__)
DRIVER_PATH = "networking_nec.plugins.openflow.drivers.%s"
DRIVER_LIST = {
    'trema': DRIVER_PATH % "trema.TremaPortBaseDriver",
    'trema_port': DRIVER_PATH % "trema.TremaPortBaseDriver",
    'trema_portmac': DRIVER_PATH % "trema.TremaPortMACBaseDriver",
    'trema_mac': DRIVER_PATH % "trema.TremaMACBaseDriver",
    'pfc': DRIVER_PATH % "pfc.PFCV51Driver",
    'pfc_v3': DRIVER_PATH % "pfc.PFCV3Driver",
    'pfc_v4': DRIVER_PATH % "pfc.PFCV4Driver",
    'pfc_v5': DRIVER_PATH % "pfc.PFCV5Driver",
    'pfc_v51': DRIVER_PATH % "pfc.PFCV51Driver",
}


def get_driver(driver_name):
    LOG.info(_LI("Loading OFC driver: %s"), driver_name)
    driver_klass = DRIVER_LIST.get(driver_name) or driver_name
    return importutils.import_class(driver_klass)



# Copyright 2012-2013 NEC Corporation.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import uuidutils
from sqlalchemy.orm import exc as sa_exc
from sqlalchemy import sql

from neutron.api.v2 import attributes

from networking_nec.plugins.openflow.db import models as nmodels
from networking_nec.plugins.openflow.extensions import packetfilter as ext_pf


PF_STATUS_ACTIVE = 'ACTIVE'
PF_STATUS_DOWN = 'DOWN'
PF_STATUS_ERROR = 'ERROR'

INT_FIELDS = ('eth_type', 'src_port', 'dst_port')


class PacketFilterDbMixin(object):

    def _make_packet_filter_dict(self, pf_entry, fields=None):
        res = {'id': pf_entry['id'],
               'name': pf_entry['name'],
               'tenant_id': pf_entry['tenant_id'],
               'network_id': pf_entry['network_id'],
               'action': pf_entry['action'],
               'priority': pf_entry['priority'],
               'in_port': pf_entry['in_port'],
               # "or None" ensure the filed is None if empty
               'src_mac': pf_entry['src_mac'] or None,
               'dst_mac': pf_entry['dst_mac'] or None,
               'eth_type': pf_entry['eth_type'] or None,
               'src_cidr': pf_entry['src_cidr'] or None,
               'dst_cidr': pf_entry['dst_cidr'] or None,
               'protocol': pf_entry['protocol'] or None,
               'src_port': pf_entry['src_port'] or None,
               'dst_port': pf_entry['dst_port'] or None,
               'admin_state_up': pf_entry['admin_state_up'],
               'status': pf_entry['status']}
        return self._fields(res, fields)

    def _get_packet_filter(self, context, id):
        try:
            pf_entry = self._get_by_id(context, nmodels.PacketFilter, id)
        except sa_exc.NoResultFound:
            raise ext_pf.PacketFilterNotFound(id=id)
        return pf_entry

    def get_packet_filter(self, context, id, fields=None):
        pf_entry = self._get_packet_filter(context, id)
        return self._make_packet_filter_dict(pf_entry, fields)

    def get_packet_filters(self, context, filters=None, fields=None):
        return self._get_collection(context,
                                    nmodels.PacketFilter,
                                    self._make_packet_filter_dict,
                                    filters=filters,
                                    fields=fields)

    def _replace_unspecified_field(self, params, key):
        if not attributes.is_attr_set(params[key]):
            if key == 'in_port':
                params[key] = None
            elif key in INT_FIELDS:
                # Integer field
                params[key] = 0
            else:
                params[key] = ''

    def _get_eth_type_for_protocol(self, protocol):
        if protocol.upper() in ("ICMP", "TCP", "UDP"):
            return 0x800
        elif protocol.upper() == "ARP":
            return 0x806

    def _set_eth_type_from_protocol(self, filter_dict):
        if filter_dict.get('protocol'):
            eth_type = self._get_eth_type_for_protocol(filter_dict['protocol'])
            if eth_type:
                filter_dict['eth_type'] = eth_type

    def _check_eth_type_and_protocol(self, new_filter, current_filter):
        if 'protocol' in new_filter or 'eth_type' not in new_filter:
            return
        eth_type = self._get_eth_type_for_protocol(current_filter['protocol'])
        if not eth_type:
            return
        if eth_type != new_filter['eth_type']:
            raise ext_pf.PacketFilterEtherTypeProtocolMismatch(
                eth_type=hex(new_filter['eth_type']),
                protocol=current_filter['protocol'])

    def create_packet_filter(self, context, packet_filter):
        pf_dict = packet_filter['packet_filter']
        tenant_id = self._get_tenant_id_for_create(context, pf_dict)

        if pf_dict['in_port'] == attributes.ATTR_NOT_SPECIFIED:
            # validate network ownership
            self.get_network(context, pf_dict['network_id'])
        else:
            # validate port ownership
            self.get_port(context, pf_dict['in_port'])

        params = {'tenant_id': tenant_id,
                  'id': pf_dict.get('id') or uuidutils.generate_uuid(),
                  'name': pf_dict['name'],
                  'network_id': pf_dict['network_id'],
                  'priority': pf_dict['priority'],
                  'action': pf_dict['action'],
                  'admin_state_up': pf_dict.get('admin_state_up', True),
                  'status': PF_STATUS_DOWN,
                  'in_port': pf_dict['in_port'],
                  'src_mac': pf_dict['src_mac'],
                  'dst_mac': pf_dict['dst_mac'],
                  'eth_type': pf_dict['eth_type'],
                  'src_cidr': pf_dict['src_cidr'],
                  'dst_cidr': pf_dict['dst_cidr'],
                  'src_port': pf_dict['src_port'],
                  'dst_port': pf_dict['dst_port'],
                  'protocol': pf_dict['protocol']}
        for key in params:
            self._replace_unspecified_field(params, key)
        self._set_eth_type_from_protocol(params)

        with context.session.begin(subtransactions=True):
            pf_entry = nmodels.PacketFilter(**params)
            context.session.add(pf_entry)

        return self._make_packet_filter_dict(pf_entry)

    def update_packet_filter(self, context, id, packet_filter):
        params = packet_filter['packet_filter']
        for key in params:
            self._replace_unspecified_field(params, key)
        self._set_eth_type_from_protocol(params)
        with context.session.begin(subtransactions=True):
            pf_entry = self._get_packet_filter(context, id)
            self._check_eth_type_and_protocol(params, pf_entry)
            pf_entry.update(params)
        return self._make_packet_filter_dict(pf_entry)

    def delete_packet_filter(self, context, id):
        with context.session.begin(subtransactions=True):
            pf_entry = self._get_packet_filter(context, id)
            context.session.delete(pf_entry)

    def get_packet_filters_for_port(self, context, port):
        """Retrieve packet filters on OFC on a given port.

        It returns a list of tuple (neutron filter_id, OFC id).
        """
        query = (context.session.query(nmodels.OFCFilterMapping)
                 .join(nmodels.PacketFilter,
                       nmodels.OFCFilterMapping.neutron_id
                       == nmodels.PacketFilter.id)
                 .filter(nmodels.PacketFilter.admin_state_up == sql.true()))

        network_id = port['network_id']
        net_pf_query = (query.filter(nmodels.PacketFilter.network_id
                                     == network_id)
                        .filter(nmodels.PacketFilter.in_port == sql.null()))
        net_filters = [(pf['neutron_id'], pf['ofc_id']) for pf in net_pf_query]

        port_pf_query = query.filter(nmodels.PacketFilter.in_port
                                     == port['id'])
        port_filters = [(pf['neutron_id'], pf['ofc_id'])
                        for pf in port_pf_query]

        return net_filters + port_filters



# Copyright 2015 NEC Corporation.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""This module implements logic to manage logical networks and ports
on an OpenFlow controller.
"""

from oslo_log import log as logging
from oslo_utils import excutils

from neutron.common import constants as const
from neutron.db import db_base_plugin_v2
from neutron.extensions import portbindings

from networking_nec._i18n import _LE, _LI, _LW
from networking_nec.plugins.openflow.db import api as ndb
from networking_nec.plugins.openflow.db import router as rdb
from networking_nec.plugins.openflow import exceptions as nexc
from networking_nec.plugins.openflow import router as router_plugin
from networking_nec.plugins.openflow import utils

LOG = logging.getLogger(__name__)


class L2Manager(object):

    def __init__(self, plugin):
        self._plugin = plugin

        self.port_handlers = {
            'create': {
                const.DEVICE_OWNER_ROUTER_GW: self.plugin.create_router_port,
                const.DEVICE_OWNER_ROUTER_INTF: self.plugin.create_router_port,
                'default': self.activate_port_if_ready,
            },
            'delete': {
                const.DEVICE_OWNER_ROUTER_GW: self.plugin.delete_router_port,
                const.DEVICE_OWNER_ROUTER_INTF: self.plugin.delete_router_port,
                'default': self.deactivate_port,
            }
        }

    @property
    def plugin(self):
        return self._plugin

    @property
    def ofc(self):
        return self._plugin.ofc

    def _check_ofc_tenant_in_use(self, context, tenant_id, deleting=None):
        """Check if the specified tenant is used."""
        # All networks are created on OFC
        filters = {'tenant_id': [tenant_id]}
        net_count = self.plugin.get_networks_count(context, filters=filters)
        if deleting == 'network':
            net_count -= 1
        if net_count:
            return True

        router_count = rdb.get_router_count_by_provider(
            context.session, router_plugin.PROVIDER_OPENFLOW, tenant_id)
        if deleting == 'router':
            router_count -= 1
        if router_count:
            return True

        return False

    def _cleanup_ofc_tenant(self, context, tenant_id, deleting):
        if not self._check_ofc_tenant_in_use(context, tenant_id, deleting):
            try:
                if self.ofc.exists_ofc_tenant(context, tenant_id):
                    self.ofc.delete_ofc_tenant(context, tenant_id)
                else:
                    LOG.debug('_cleanup_ofc_tenant: No OFC tenant for %s',
                              tenant_id)
            except (nexc.OFCException, nexc.OFCMappingNotFound) as exc:
                LOG.warn(_LW("delete_ofc_tenant() failed due to %s"), exc)

    def activate_port_if_ready(self, context, port, network=None):
        """Activate port by creating port on OFC if ready.

        Conditions to activate port on OFC are:
            * port admin_state is UP
            * network admin_state is UP
            * portinfo are available (to identify port on OFC)
        """
        if not network:
            network = self.plugin.get_network(context, port['network_id'])

        if not port['admin_state_up']:
            LOG.debug("activate_port_if_ready(): skip, "
                      "port.admin_state_up is False.")
            return port
        elif not network['admin_state_up']:
            LOG.debug("activate_port_if_ready(): skip, "
                      "network.admin_state_up is False.")
            return port
        elif not ndb.get_portinfo(context.session, port['id']):
            LOG.debug("activate_port_if_ready(): skip, "
                      "no portinfo for this port.")
            return port
        elif self.ofc.exists_ofc_port(context, port['id']):
            LOG.debug("activate_port_if_ready(): skip, "
                      "ofc_port already exists.")
            return port

        try:
            self.ofc.create_ofc_port(context, port['id'], port)
            port_status = const.PORT_STATUS_ACTIVE
        except (nexc.OFCException, nexc.OFCMappingNotFound) as exc:
            LOG.error(_LE("create_ofc_port() failed due to %s"), exc)
            port_status = const.PORT_STATUS_ERROR

        if port_status != port['status']:
            utils.update_resource_status(context, "port", port['id'],
                                         port_status)
            port['status'] = port_status

        return port

    def deactivate_port(self, context, port, raise_exc=True):
        """Deactivate port by deleting port from OFC if exists."""
        if not self.ofc.exists_ofc_port(context, port['id']):
            LOG.debug("deactivate_port(): skip, ofc_port for port=%s "
                      "does not exist.", port['id'])
            return port

        try:
            self.ofc.delete_ofc_port(context, port['id'], port)
            utils.update_resource_status_if_changed(
                context, "port", port, const.PORT_STATUS_DOWN,
                ignore_error=True)
            return port
        except (nexc.OFCResourceNotFound, nexc.OFCMappingNotFound):
            # There is a case where multiple delete_port operation are
            # running concurrently. For example, delete_port from
            # release_dhcp_port and deletion of network owned ports in
            # delete_network. In such cases delete_ofc_port may receive
            # 404 error from OFC.
            # Also there is a case where neutron port is deleted
            # between exists_ofc_port and get_ofc_id in delete_ofc_port.
            # In this case OFCMappingNotFound is raised.
            # These two cases are valid situations.
            LOG.info(_LI("deactivate_port(): OFC port for port=%s is "
                         "already removed."), port['id'])
            # The port is already removed, so there is no need
            # to update status in the database.
            port['status'] = const.PORT_STATUS_DOWN
            return port
        except nexc.OFCException as exc:
            with excutils.save_and_reraise_exception() as ctxt:
                LOG.error(_LE("Failed to delete port=%(port)s from OFC: "
                              "%(exc)s"), {'port': port['id'], 'exc': exc})
                utils.update_resource_status_if_changed(
                    context, "port", port, const.PORT_STATUS_ERROR,
                    ignore_error=True)
                if not raise_exc:
                    ctxt.reraise = False
                    return port

    def _net_status(self, network):
        # NOTE: NEC Plugin accept admin_state_up. When it's False, this plugin
        # deactivate all ports on the network to drop all packet and show
        # status='DOWN' to users. But the network is kept defined on OFC.
        if network['admin_state_up']:
            return const.NET_STATUS_ACTIVE
        else:
            return const.NET_STATUS_DOWN

    def get_initial_net_status(self, network):
        return self._net_status(network['network'])
        # TODO(amotoki): Set the network initial status to BUILD and
        # update it to ACTIVE/ERROR after creating a network on OFC.
        # return const.NET_STATUS_BUILD

    def get_initial_port_status(self, port):
        # NOTE(amotoki): Not sure the initial port status should be
        # BUILD or DOWN. Both works with the current nova/neutron.
        if port['port'].get('status'):
            return port['port'].get('status')
        return const.PORT_STATUS_DOWN

    def create_network(self, context, network):
        tenant_id = network['tenant_id']
        net_id = network['id']
        net_name = network['name']
        try:
            if not self.ofc.exists_ofc_tenant(context, tenant_id):
                self.ofc.create_ofc_tenant(context, tenant_id)
            self.ofc.create_ofc_network(context, tenant_id, net_id, net_name)
            utils.update_resource_status_if_changed(
                context, "network", network, self._net_status(network))
        except (nexc.OFCException, nexc.OFCMappingNotFound) as exc:
            LOG.error(_LE("Failed to create network id=%(id)s on "
                          "OFC: %(exc)s"), {'id': net_id, 'exc': exc})
            utils.update_resource_status_if_changed(
                context, "network", network, const.NET_STATUS_ERROR)

    def update_network(self, context, id, old_net, new_net):
        changed = (old_net['admin_state_up'] != new_net['admin_state_up'])
        if changed:
            new_status = self._net_status(new_net)
            utils.update_resource_status_if_changed(
                context, "network", new_net, new_status)
        if changed and not new_net['admin_state_up']:
            # disable all active ports of the network
            filters = dict(network_id=[id], status=[const.PORT_STATUS_ACTIVE])
            ports = self.plugin.get_ports(context, filters=filters)
            for port in ports:
                # If some error occurs, status of errored port is set to ERROR.
                # This is avoids too many rollback.
                # TODO(amotoki): Raise an exception after all port operations
                # are finished to inform the caller of API of the failure.
                self.deactivate_port(context, port, raise_exc=False)
        elif changed and new_net['admin_state_up']:
            # enable ports of the network
            filters = dict(network_id=[id], status=[const.PORT_STATUS_DOWN],
                           admin_state_up=[True])
            ports = self.plugin.get_ports(context, filters=filters)
            for port in ports:
                self.activate_port_if_ready(context, port, new_net)

    def delete_network(self, context, id, ports=None):
        net_db = self.plugin._get_network(context, id)
        tenant_id = net_db['tenant_id']

        if ports is None:
            ports = self.plugin.get_ports(context,
                                          filters={'network_id': [id]})

        # Make sure auto-delete ports on OFC are deleted.
        # If an error occurs during port deletion,
        # delete_network will be aborted.
        for port in [p for p in ports if p['device_owner']
                     in db_base_plugin_v2.AUTO_DELETE_PORT_OWNERS]:
            port = self.deactivate_port(context, port)

        # delete all packet_filters of the network from the controller
        for pf in net_db.packetfilters:
            self.plugin.delete_packet_filter(context, pf['id'])

        if self.ofc.exists_ofc_network(context, id):
            try:
                self.ofc.delete_ofc_network(context, id, net_db)
            except (nexc.OFCException, nexc.OFCMappingNotFound) as exc:
                with excutils.save_and_reraise_exception():
                    LOG.error(_LE("delete_network() failed due to %s"), exc)
                    utils.update_resource_status(
                        context, "network", net_db['id'],
                        const.NET_STATUS_ERROR)

        self._cleanup_ofc_tenant(context, tenant_id, deleting='network')

    def _get_port_handler(self, operation, device_owner):
        handlers = self.port_handlers[operation]
        handler = handlers.get(device_owner)
        if handler:
            return handler
        else:
            return handlers['default']

    def create_port(self, context, port):
        handler = self._get_port_handler('create', port['device_owner'])
        return handler(context, port)

    @staticmethod
    def get_portinfo(port):
        profile = port.get(portbindings.PROFILE)
        if not profile:
            return
        return {'datapath_id': profile['datapath_id'],
                'port_no': profile['port_no']}

    @staticmethod
    def is_portinfo_changed(old_port, new_port):
        """Check portinfo is changed or not.

        :param old_port: old port information
        :param new_port: new port information
        :returns: 'ADD', 'MOD', 'DEL' or None
        """
        old_portinfo = L2Manager.get_portinfo(old_port)
        new_portinfo = L2Manager.get_portinfo(new_port)

        # portinfo has been validated, so we can assume
        # portinfo is either None or a valid dict.
        if not old_portinfo and not new_portinfo:
            return
        elif old_portinfo and not new_portinfo:
            return 'DEL'
        elif not old_portinfo and new_portinfo:
            return 'ADD'
        else:
            if (utils.cmp_dpid(old_portinfo['datapath_id'],
                               new_portinfo['datapath_id']) and
                    old_portinfo['port_no'] == new_portinfo['port_no']):
                return
            else:
                return 'MOD'

    @staticmethod
    def get_ofport_exist(port):
        return (port['admin_state_up'] and
                bool(port.get(portbindings.PROFILE)))

    def _update_ofc_port_if_required(self, context, old_port, new_port):
        # Determine it is required to update OFC port
        need_add = False
        need_del = False
        need_packet_filter_update = False

        old_ofport_exist = self.get_ofport_exist(old_port)
        new_ofport_exist = self.get_ofport_exist(new_port)
        portinfo_changed = self.is_portinfo_changed(old_port, new_port)

        if old_port['admin_state_up'] != new_port['admin_state_up']:
            if new_port['admin_state_up']:
                need_add |= new_ofport_exist
            else:
                need_del |= old_ofport_exist

        if portinfo_changed:
            if portinfo_changed in ['DEL', 'MOD']:
                need_del |= old_ofport_exist
            if portinfo_changed in ['ADD', 'MOD']:
                need_add |= new_ofport_exist
            need_packet_filter_update |= True

        # Update OFC port if required
        if need_del:
            self.deactivate_port(context, new_port)
            if need_packet_filter_update:
                self.plugin.deactivate_packet_filters_by_port(context, id)
        if need_add:
            if need_packet_filter_update:
                self.plugin.activate_packet_filters_by_port(context, id)
            self.activate_port_if_ready(context, new_port)

    def update_port(self, context, old_port, new_port):
        self._update_ofc_port_if_required(context, old_port, new_port)
        return new_port

    def delete_port(self, context, id):
        # ext_sg.SECURITYGROUPS attribute for the port is required
        # since notifier.security_groups_member_updated() need the attribute.
        # Thus we need to call self.get_port() instead of super().get_port()
        port_db = self.plugin._get_port(context, id)
        port = self.plugin._make_port_dict(port_db)

        handler = self._get_port_handler('delete', port['device_owner'])
        # handler() raises an exception if an error occurs during processing.
        port = handler(context, port)

        # delete all packet_filters of the port from the controller
        for pf in port_db['packetfilters']:
            self.plugin.delete_packet_filter(context, pf['id'])

        return port



# Copyright 2013 NEC Corporation.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from neutron.api.rpc.handlers import l3_rpc
from neutron.common import constants
from neutron.tests.unit.db import test_agentschedulers_db \
    as test_agent_scheduler

from networking_nec.tests.unit.openflow import test_plugin

L3_HOSTA = test_agent_scheduler.L3_HOSTA
L3_HOSTB = test_agent_scheduler.L3_HOSTB


class NecAgentSchedulerTestCase(
        test_agent_scheduler.OvsAgentSchedulerTestCase,
        test_plugin.NecPluginV2TestCaseBase):

    plugin_str = test_plugin.PLUGIN_NAME
    l3_plugin = None

    def setUp(self):
        if '_dvr_' in self._testMethodName:
            self.skipTest("NEC OpenFlow Plugin does not support "
                          "DVR agent scheduler.")
        self.setup_nec_plugin_base()
        super(NecAgentSchedulerTestCase, self).setUp()


class NecDhcpAgentNotifierTestCase(
        test_agent_scheduler.OvsDhcpAgentNotifierTestCase,
        test_plugin.NecPluginV2TestCaseBase):

    plugin_str = test_plugin.PLUGIN_NAME
    l3_plugin = None

    def setUp(self):
        self.setup_nec_plugin_base()
        super(NecDhcpAgentNotifierTestCase, self).setUp()


class NecL3AgentNotifierTestCase(
        test_agent_scheduler.OvsL3AgentNotifierTestCase,
        test_plugin.NecPluginV2TestCaseBase):

    plugin_str = test_plugin.PLUGIN_NAME
    l3_plugin = None

    def setUp(self):
        self.setup_nec_plugin_base()
        super(NecL3AgentNotifierTestCase, self).setUp()


class NecL3AgentSchedulerWithOpenFlowRouter(
        test_agent_scheduler.OvsAgentSchedulerTestCaseBase,
        test_plugin.NecPluginV2TestCaseBase):

    plugin_str = test_plugin.PLUGIN_NAME
    l3_plugin = None

    def setUp(self):
        self.setup_nec_plugin_base()
        super(NecL3AgentSchedulerWithOpenFlowRouter, self).setUp()

    def test_router_auto_schedule_with_l3agent_and_openflow(self):
        with self.router() as r1, \
                self.router(arg_list=('provider',),
                            provider='openflow') as r2:
            l3_rpc_cb = l3_rpc.L3RpcCallback()
            self._register_agent_states()
            ret_a = l3_rpc_cb.sync_routers(self.adminContext, host=L3_HOSTA)
            ret_b = l3_rpc_cb.sync_routers(self.adminContext, host=L3_HOSTB)
            l3_agents = self._list_l3_agents_hosting_router(
                r1['router']['id'])
            self.assertEqual(1, len(ret_a))
            self.assertFalse(len(ret_b))
            self.assertIn(r1['router']['id'], [r['id'] for r in ret_a])
            self.assertNotIn(r2['router']['id'], [r['id'] for r in ret_a])
        self.assertEqual(1, len(l3_agents['agents']))
        self.assertEqual(L3_HOSTA, l3_agents['agents'][0]['host'])

    def test_router_auto_schedule_only_with_openflow_router(self):
        with self.router(arg_list=('provider',), provider='openflow') as r1, \
                self.router(arg_list=('provider',), provider='openflow') as r2:
            l3_rpc_cb = l3_rpc.L3RpcCallback()
            self._register_agent_states()
            ret_a = l3_rpc_cb.sync_routers(self.adminContext, host=L3_HOSTA)
            l3_agents_1 = self._list_l3_agents_hosting_router(
                r1['router']['id'])
            l3_agents_2 = self._list_l3_agents_hosting_router(
                r2['router']['id'])
            self.assertFalse(len(ret_a))
            self.assertNotIn(r1['router']['id'], [r['id'] for r in ret_a])
            self.assertNotIn(r2['router']['id'], [r['id'] for r in ret_a])
        self.assertFalse(len(l3_agents_1['agents']))
        self.assertFalse(len(l3_agents_2['agents']))

    def test_add_router_to_l3_agent_for_openflow_router(self):
        with self.router(arg_list=('provider',), provider='openflow') as r1:
            self._register_agent_states()
            hosta_id = self._get_agent_id(constants.AGENT_TYPE_L3,
                                          L3_HOSTA)
            self._add_router_to_l3_agent(hosta_id,
                                         r1['router']['id'],
                                         expected_code=409)



# Copyright 2014 NEC Corporation.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from sqlalchemy.orm import exc as sa_exc

from neutron.common import constants
from neutron import manager
from neutron.tests import base

from networking_nec.plugins.openflow import utils
from networking_nec.tests.unit.openflow import test_plugin


class NecUtilsTest(base.BaseTestCase):

    def test_cmp_dpid(self):
        self.assertTrue(utils.cmp_dpid('0xabcd', '0xabcd'))
        self.assertTrue(utils.cmp_dpid('abcd', '0xabcd'))
        self.assertTrue(utils.cmp_dpid('0x000000000000abcd', '0xabcd'))
        self.assertTrue(utils.cmp_dpid('0x000000000000abcd', '0x00abcd'))
        self.assertFalse(utils.cmp_dpid('0x000000000000abcd', '0xabc0'))
        self.assertFalse(utils.cmp_dpid('0x000000000000abcd', '0x00abc0'))

    def test_cmp_dpid_with_exception(self):
        self.assertFalse(utils.cmp_dpid('0xabcx', '0xabcx'))
        self.assertFalse(utils.cmp_dpid(None, None))


class TestUtilsDbTest(test_plugin.NecPluginV2TestCase):

    def test_update_resource(self):
        with self.network() as network:
            self.assertEqual("ACTIVE", network['network']['status'])
            net_id = network['network']['id']
            for status in ["DOWN", "BUILD", "ERROR", "ACTIVE"]:
                utils.update_resource_status(
                    self.context, 'network', net_id,
                    getattr(constants, 'NET_STATUS_%s' % status))
                n = self.plugin._get_network(self.context, net_id)
                self.assertEqual(status, n.status)

    def _test_update_resource_with_exception(self, exc=None,
                                             ignore_error=False,
                                             exc_expexted=True):
        def _call_method():
            utils.update_resource_status(
                self.context, 'network', net_id, "DOWN", ignore_error)

        with self.network() as network:
            self.assertEqual("ACTIVE", network['network']['status'])
            net_id = network['network']['id']
            plugin = manager.NeutronManager.get_plugin()
            exc = exc or Exception
            with mock.patch.object(plugin, '_get_network', side_effect=exc):
                if exc_expexted:
                    self.assertRaises(exc, _call_method)
                else:
                    _call_method()
            n = self.plugin._get_network(self.context, net_id)
            # Check the status is unchanged.
            self.assertEqual("ACTIVE", n.status)

    def test_update_resource_with_staledata_error_ignore_error(self):
        self._test_update_resource_with_exception(sa_exc.StaleDataError,
                                                  ignore_error=True,
                                                  exc_expexted=False)

    def test_update_resource_with_staledata_error(self):
        self._test_update_resource_with_exception(sa_exc.StaleDataError,
                                                  ignore_error=False)

    def test_update_resource_with_general_exception_with_ignore_error(self):
        self._test_update_resource_with_exception(ignore_error=True)

    def test_update_resource_with_general_exception(self):
        self._test_update_resource_with_exception(ignore_error=False)



# Copyright 2012 NEC Corporation.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr
from oslo_log import helpers as log_helpers
from oslo_log import log as logging
from oslo_utils import uuidutils

from networking_nec.plugins.openflow import exceptions as nexc
from networking_nec.plugins.openflow import ofc_driver_base


LOG = logging.getLogger(__name__)

MAX_NUM_OPENFLOW_ROUTER = 2


class StubOFCDriver(ofc_driver_base.OFCDriverBase):
    """Stub OFC driver for testing.

    This driver can be used not only for unit tests but also for real testing
    as a logging driver. It stores the created resources on OFC and returns
    them in get methods().

    If autocheck is enabled, it checks whether the specified resource exists
    in OFC and raises an exception if it is different from expected status.
    """

    def __init__(self, conf):
        self.autocheck = False
        self.reset_all()

    def reset_all(self):
        self.ofc_tenant_dict = {}
        self.ofc_network_dict = {}
        self.ofc_port_dict = {}
        self.ofc_filter_dict = {}
        self.ofc_router_dict = {}
        self.ofc_router_inf_dict = {}
        self.ofc_router_route_dict = {}

    def enable_autocheck(self):
        self.autocheck = True

    def disable_autocheck(self):
        self.autocheck = False

    @log_helpers.log_method_call
    def create_tenant(self, description, tenant_id=None):
        ofc_id = "ofc-" + tenant_id[:-4]
        if self.autocheck:
            if ofc_id in self.ofc_tenant_dict:
                raise Exception(_('(create_tenant) OFC tenant %s '
                                  'already exists') % ofc_id)
        self.ofc_tenant_dict[ofc_id] = {'tenant_id': tenant_id,
                                        'description': description}
        return ofc_id

    @log_helpers.log_method_call
    def delete_tenant(self, ofc_tenant_id):
        if ofc_tenant_id in self.ofc_tenant_dict:
            del self.ofc_tenant_dict[ofc_tenant_id]
        else:
            if self.autocheck:
                raise Exception(_('(delete_tenant) OFC tenant %s not found')
                                % ofc_tenant_id)
        LOG.debug('delete_tenant: SUCCEED')

    @log_helpers.log_method_call
    def create_network(self, ofc_tenant_id, description, network_id=None):
        ofc_id = "ofc-" + network_id[:-4]
        if self.autocheck:
            if ofc_tenant_id not in self.ofc_tenant_dict:
                raise Exception(_('(create_network) OFC tenant %s not found')
                                % ofc_tenant_id)
            if ofc_id in self.ofc_network_dict:
                raise Exception(_('(create_network) OFC network %s '
                                  'already exists') % ofc_id)
        self.ofc_network_dict[ofc_id] = {'tenant_id': ofc_tenant_id,
                                         'network_id': network_id,
                                         'description': description}
        return ofc_id

    @log_helpers.log_method_call
    def update_network(self, ofc_network_id, description):
        if self.autocheck:
            if ofc_network_id not in self.ofc_network_dict:
                raise Exception(_('(update_network) OFC network %s not found')
                                % ofc_network_id)
        data = {'description': description}
        self.ofc_network_dict[ofc_network_id].update(data)
        LOG.debug('update_network: SUCCEED')

    @log_helpers.log_method_call
    def delete_network(self, ofc_network_id):
        if ofc_network_id in self.ofc_network_dict:
            del self.ofc_network_dict[ofc_network_id]
        else:
            if self.autocheck:
                raise Exception(_('(delete_network) OFC network %s not found')
                                % ofc_network_id)
        LOG.debug('delete_network: SUCCEED')

    @log_helpers.log_method_call
    def create_port(self, ofc_network_id, info, port_id=None, filters=None):
        ofc_id = "ofc-" + port_id[:-4]
        if self.autocheck:
            if ofc_network_id not in self.ofc_network_dict:
                raise Exception(_('(create_port) OFC network %s not found')
                                % ofc_network_id)
            if ofc_id in self.ofc_port_dict:
                raise Exception(_('(create_port) OFC port %s already exists')
                                % ofc_id)
        self.ofc_port_dict[ofc_id] = {'network_id': ofc_network_id,
                                      'port_id': port_id}
        if filters:
            self.ofc_port_dict[ofc_id]['filters'] = filters
        return ofc_id

    @log_helpers.log_method_call
    def delete_port(self, ofc_port_id):
        if ofc_port_id in self.ofc_port_dict:
            del self.ofc_port_dict[ofc_port_id]
        else:
            if self.autocheck:
                raise Exception(_('(delete_port) OFC port %s not found')
                                % ofc_port_id)
        LOG.debug('delete_port: SUCCEED')

    @classmethod
    def filter_supported(cls):
        return True

    def create_filter(self, context, filter_dict, filter_id=None):
        return "ofc-" + filter_id[:-4]

    def delete_filter(self, ofc_filter_id):
        pass

    def convert_ofc_tenant_id(self, context, ofc_tenant_id):
        return ofc_tenant_id

    def convert_ofc_network_id(self, context, ofc_network_id, tenant_id):
        return ofc_network_id

    def convert_ofc_port_id(self, context, ofc_port_id, tenant_id, network_id):
        return ofc_port_id

    def convert_ofc_filter_id(self, context, ofc_filter_id):
        return ofc_filter_id

    router_supported = True
    router_nat_supported = True

    @log_helpers.log_method_call
    def create_router(self, ofc_tenant_id, router_id, description):
        ofc_id = "ofc-" + router_id[:-4]
        if self.autocheck:
            if ofc_tenant_id not in self.ofc_tenant_dict:
                raise Exception(_('(create_router) OFC tenant %s not found')
                                % ofc_tenant_id)
            if ofc_id in self.ofc_router_dict:
                raise Exception(_('(create_router) OFC router %s '
                                  'already exists') % ofc_id)
        if len(self.ofc_router_dict) >= MAX_NUM_OPENFLOW_ROUTER:
            params = {'reason': _("Operation on OFC is failed"),
                      'status': 409}
            raise nexc.OFCException(**params)
        self.ofc_router_dict[ofc_id] = {'tenant_id': ofc_tenant_id,
                                        'router_id': router_id,
                                        'description': description}
        return ofc_id

    @log_helpers.log_method_call
    def delete_router(self, ofc_router_id):
        if ofc_router_id in self.ofc_router_dict:
            del self.ofc_router_dict[ofc_router_id]
        else:
            if self.autocheck:
                raise Exception(_('(delete_router) OFC router %s not found')
                                % ofc_router_id)
        LOG.debug('delete_router: SUCCEED')

    @log_helpers.log_method_call
    def add_router_interface(self, ofc_router_id, ofc_net_id,
                             ip_address=None, mac_address=None):
        if_id = "ofc-" + uuidutils.generate_uuid()[:-4]
        # IP address should have a format of a.b.c.d/N
        if ip_address != str(netaddr.IPNetwork(ip_address)):
            raise Exception(_('(add_router_interface) '
                            'ip_address %s is not a valid format (a.b.c.d/N).')
                            % ip_address)
        if self.autocheck:
            if ofc_router_id not in self.ofc_router_dict:
                raise Exception(_('(add_router_interface) '
                                  'OFC router %s not found') % ofc_router_id)
            if ofc_net_id not in self.ofc_network_dict:
                raise Exception(_('(add_router_interface) '
                                  'OFC network %s not found') % ofc_net_id)
            # Check duplicate destination
        self.ofc_router_inf_dict[if_id] = {'router_id': ofc_router_id,
                                           'network_id': ofc_net_id,
                                           'ip_address': ip_address,
                                           'mac_address': mac_address}
        LOG.debug('add_router_interface: SUCCEED (if_id=%s)', if_id)
        return if_id

    @log_helpers.log_method_call
    def update_router_interface(self, ofc_router_inf_id,
                                ip_address=None, mac_address=None):
        if ofc_router_inf_id not in self.ofc_router_inf_dict:
            if self.autocheck:
                raise Exception(_('(delete_router_interface) '
                                  'OFC router interface %s not found')
                                % ofc_router_inf_id)
            self.ofc_router_inf_dict[ofc_router_inf_id] = {}
        inf = self.ofc_router_inf_dict[ofc_router_inf_id]
        if ip_address:
            inf.update({'ip_address': ip_address})
        if mac_address:
            inf.update({'mac_address': mac_address})
        LOG.debug('update_router_route: SUCCEED')

    @log_helpers.log_method_call
    def delete_router_interface(self, ofc_router_inf_id):
        if ofc_router_inf_id in self.ofc_router_inf_dict:
            del self.ofc_router_inf_dict[ofc_router_inf_id]
        else:
            if self.autocheck:
                raise Exception(_('(delete_router_interface) '
                                  'OFC router interface %s not found')
                                % ofc_router_inf_id)
        LOG.debug('delete_router_interface: SUCCEED')

    @log_helpers.log_method_call
    def add_router_route(self, ofc_router_id, destination, nexthop):
        route_id = "ofc-" + uuidutils.generate_uuid()[:-4]
        # IP address format check
        netaddr.IPNetwork(destination)
        netaddr.IPAddress(nexthop)
        if self.autocheck:
            if ofc_router_id not in self.ofc_router_dict:
                raise Exception(_('(add_router_route) OFC router %s not found')
                                % ofc_router_id)
            # Check duplicate destination
            if destination in [route['destination'] for route in
                               self.ofc_router_route_dict.values()]:
                raise Exception(_('(add_router_route) '
                                'route to "%s" already exists') % destination)
        self.ofc_router_route_dict[route_id] = {'router_id': ofc_router_id,
                                                'destination': destination,
                                                'nexthop': nexthop}
        LOG.debug('add_router_route: SUCCEED (route_id=%s)', route_id)
        return route_id

    @log_helpers.log_method_call
    def delete_router_route(self, ofc_router_route_id):
        if ofc_router_route_id in self.ofc_router_route_dict:
            del self.ofc_router_route_dict[ofc_router_route_id]
        else:
            if self.autocheck:
                raise Exception(_('(delete_router_route) OFC router route %s '
                                  'not found') % ofc_router_route_id)
        LOG.debug('delete_router_route: SUCCEED')

    @log_helpers.log_method_call
    def list_router_routes(self, ofc_router_id):
        if self.autocheck:
            if ofc_router_id not in self.ofc_router_dict:
                raise Exception(_('(delete_router) OFC router %s not found')
                                % ofc_router_id)
        routes = [{'id': k,
                   'destination': v['destination'],
                   'nexthop': v['nexthop']}
                  for k, v in self.ofc_router_route_dict.items()
                  if v['router_id'] == ofc_router_id]
        LOG.debug('list_router_routes: routes=%s', routes)
        return routes



# Copyright 2013 NEC Corporation.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils
from oslo_utils import importutils

from neutron.api.rpc.agentnotifiers import l3_rpc_agent_api
from neutron.api.v2 import attributes as attr
from neutron.common import exceptions as n_exc
from neutron.db import db_base_plugin_v2
from neutron.db import extraroute_db
from neutron.db import l3_agentschedulers_db
from neutron.db import l3_db
from neutron.db import l3_gwmode_db
from neutron.db import models_v2
from neutron.extensions import l3

from networking_nec._i18n import _LE, _LI, _LW
from networking_nec.plugins.openflow import constants as nconst
from networking_nec.plugins.openflow.db import router as rdb
from networking_nec.plugins.openflow import exceptions as nexc
from networking_nec.plugins.openflow.extensions import router_provider

LOG = logging.getLogger(__name__)

PROVIDER_L3AGENT = nconst.ROUTER_PROVIDER_L3AGENT
PROVIDER_OPENFLOW = nconst.ROUTER_PROVIDER_OPENFLOW

ROUTER_DRIVER_PATH = 'networking_nec.plugins.openflow.router_drivers.'
ROUTER_DRIVER_MAP = {
    PROVIDER_L3AGENT: ROUTER_DRIVER_PATH + 'RouterL3AgentDriver',
    PROVIDER_OPENFLOW: ROUTER_DRIVER_PATH + 'RouterOpenFlowDriver'
}

ROUTER_DRIVERS = {}

STATUS_ACTIVE = nconst.ROUTER_STATUS_ACTIVE
STATUS_ERROR = nconst.ROUTER_STATUS_ERROR


class RouterMixin(extraroute_db.ExtraRoute_db_mixin,
                  l3_gwmode_db.L3_NAT_db_mixin):

    def create_router(self, context, router):
        """Create a new router entry on DB, and create it on OFC."""
        LOG.debug("RouterMixin.create_router() called, "
                  "router=%s .", router)
        tenant_id = self._get_tenant_id_for_create(context, router['router'])

        provider = get_provider_with_default(
            router['router'].get(router_provider.ROUTER_PROVIDER))
        driver = get_driver_by_provider(provider)

        with context.session.begin(subtransactions=True):
            new_router = super(RouterMixin, self).create_router(context,
                                                                router)
            new_router['gw_port'] = self._get_gw_port_detail(
                context, driver, new_router['gw_port_id'])
            rdb.add_router_provider_binding(context.session,
                                            provider, str(new_router['id']))
            self._extend_router_dict_provider(new_router, provider)

        # create router on the network controller
        try:
            return driver.create_router(context, tenant_id, new_router)
        except nexc.RouterOverLimit:
            with excutils.save_and_reraise_exception():
                super(RouterMixin, self).delete_router(context,
                                                       new_router['id'])

    def update_router(self, context, router_id, router):
        LOG.debug("RouterMixin.update_router() called, "
                  "id=%(id)s, router=%(router)s .",
                  {'id': router_id, 'router': router})

        with context.session.begin(subtransactions=True):
            old_rtr = super(RouterMixin, self).get_router(context, router_id)
            provider = old_rtr[router_provider.ROUTER_PROVIDER]
            driver = get_driver_by_provider(provider)
            old_rtr['gw_port'] = self._get_gw_port_detail(
                context, driver, old_rtr['gw_port_id'])
            new_rtr = super(RouterMixin, self).update_router(
                context, router_id, router)
            new_rtr['gw_port'] = self._get_gw_port_detail(
                context, driver, new_rtr['gw_port_id'])
        driver.update_router(context, router_id, old_rtr, new_rtr)
        return new_rtr

    def delete_router(self, context, router_id):
        LOG.debug("RouterMixin.delete_router() called, id=%s.", router_id)

        router = super(RouterMixin, self).get_router(context, router_id)
        tenant_id = router['tenant_id']
        # Since l3_db.delete_router() has no interaction with the plugin layer,
        # we need to check if the router can be deleted first.
        self._check_router_in_use(context, router_id)
        driver = self._get_router_driver_by_id(context, router_id)
        # If gw_port exists, remove it.
        gw_port = self._get_gw_port(context, router_id)
        if gw_port:
            driver.delete_interface(context, router_id, gw_port)
        driver.delete_router(context, router_id, router)

        super(RouterMixin, self).delete_router(context, router_id)

        self.l2mgr._cleanup_ofc_tenant(context, tenant_id, deleting='router')

    def add_router_interface(self, context, router_id, interface_info):
        LOG.debug("RouterMixin.add_router_interface() called, "
                  "id=%(id)s, interface=%(interface)s.",
                  {'id': router_id, 'interface': interface_info})
        return super(RouterMixin, self).add_router_interface(
            context, router_id, interface_info)

    def remove_router_interface(self, context, router_id, interface_info):
        LOG.debug("RouterMixin.remove_router_interface() called, "
                  "id=%(id)s, interface=%(interface)s.",
                  {'id': router_id, 'interface': interface_info})
        return super(RouterMixin, self).remove_router_interface(
            context, router_id, interface_info)

    def create_router_port(self, context, port):
        # This method is called from plugin.create_port()
        router_id = port['device_id']
        driver = self._get_router_driver_by_id(context, router_id)
        port = driver.add_interface(context, router_id, port)
        return port

    def delete_router_port(self, context, port):
        # This method is called from plugin.delete_port()
        router_id = port['device_id']
        driver = self._get_router_driver_by_id(context, router_id)
        return driver.delete_interface(context, router_id, port)

    def _get_gw_port_detail(self, context, driver, gw_port_id):
        if not gw_port_id or not driver.need_gw_info:
            return
        ctx_elevated = context.elevated()
        gw_port = self._get_port(ctx_elevated, gw_port_id)
        # At this moment gw_port has been created, so it is guaranteed
        # that fixed_ip is assigned for the gw_port.
        ext_subnet_id = gw_port['fixed_ips'][0]['subnet_id']
        ext_subnet = self._get_subnet(ctx_elevated, ext_subnet_id)
        gw_info = {'network_id': gw_port['network_id'],
                   'ip_address': gw_port['fixed_ips'][0]['ip_address'],
                   'mac_address': gw_port['mac_address'],
                   'cidr': ext_subnet['cidr'],
                   'gateway_ip': ext_subnet['gateway_ip']}
        return gw_info

    def _get_gw_port(self, context, router_id):
        device_filter = {'device_id': [router_id],
                         'device_owner': [l3_db.DEVICE_OWNER_ROUTER_GW]}
        ports = self.get_ports(context.elevated(), filters=device_filter)
        if ports:
            return ports[0]

    def _check_router_in_use(self, context, router_id):
        with context.session.begin(subtransactions=True):
            # Ensure that the router is not used
            router_filter = {'router_id': [router_id]}
            fips = self.get_floatingips_count(context.elevated(),
                                              filters=router_filter)
            if fips:
                raise l3.RouterInUse(router_id=router_id)

            device_filter = {'device_id': [router_id],
                             'device_owner': [l3_db.DEVICE_OWNER_ROUTER_INTF]}
            ports = self.get_ports_count(context.elevated(),
                                         filters=device_filter)
            if ports:
                raise l3.RouterInUse(router_id=router_id)

    def _get_router_for_floatingip(self, context, internal_port,
                                   internal_subnet_id,
                                   external_network_id):
        """Get a router for a requested floating IP.

        OpenFlow vrouter does not support NAT, so we need to exclude them
        from candidate routers for floating IP association.
        This method is called in l3_db._get_assoc_data().
        """
        subnet_db = self._get_subnet(context, internal_subnet_id)
        if not subnet_db['gateway_ip']:
            msg = (_('Cannot add floating IP to port on subnet %s '
                     'which has no gateway_ip') % internal_subnet_id)
            raise n_exc.BadRequest(resource='floatingip', msg=msg)

        # find router interface ports on this network
        router_intf_qry = context.session.query(models_v2.Port)
        router_intf_ports = router_intf_qry.filter_by(
            network_id=internal_port['network_id'],
            device_owner=l3_db.DEVICE_OWNER_ROUTER_INTF)

        for intf_p in router_intf_ports:
            if intf_p['fixed_ips'][0]['subnet_id'] == internal_subnet_id:
                router_id = intf_p['device_id']
                router_gw_qry = context.session.query(models_v2.Port)
                has_gw_port = router_gw_qry.filter_by(
                    network_id=external_network_id,
                    device_id=router_id,
                    device_owner=l3_db.DEVICE_OWNER_ROUTER_GW).count()
                driver = self._get_router_driver_by_id(context, router_id)
                if (has_gw_port and driver.floating_ip_support()):
                    return router_id

        raise l3.ExternalGatewayForFloatingIPNotFound(
            subnet_id=internal_subnet_id,
            external_network_id=external_network_id,
            port_id=internal_port['id'])

    def _get_sync_routers(self, context, router_ids=None, active=None):
        """Query routers and their gw ports for l3 agent.

        The difference from the superclass in l3_db is that this method
        only lists routers hosted on l3-agents.
        """
        router_list = super(RouterMixin, self)._get_sync_routers(
            context, router_ids, active)
        if router_list:
            _router_ids = [r['id'] for r in router_list]
            agent_routers = rdb.get_routers_by_provider(
                context.session, 'l3-agent',
                router_ids=_router_ids)
            router_list = [r for r in router_list
                           if r['id'] in agent_routers]
        return router_list

    def _get_router_driver_by_id(self, context, router_id):
        provider = self._get_provider_by_router_id(context, router_id)
        return get_driver_by_provider(provider)

    def _get_provider_by_router_id(self, context, router_id):
        return rdb.get_provider_by_router(context.session, router_id)

    def _extend_router_dict_provider(self, router_res, provider):
        router_res[router_provider.ROUTER_PROVIDER] = provider

    def extend_router_dict_provider(self, router_res, router_db):
        # NOTE: router_db.provider is None just after creating a router,
        # so we need to skip setting router_provider here.
        if not router_db.provider:
            return
        self._extend_router_dict_provider(router_res,
                                          router_db.provider['provider'])

    db_base_plugin_v2.NeutronDbPluginV2.register_dict_extend_funcs(
        l3.ROUTERS, [extend_router_dict_provider])


class L3AgentSchedulerDbMixin(l3_agentschedulers_db.L3AgentSchedulerDbMixin):

    def auto_schedule_routers(self, context, host, router_ids):
        router_ids = rdb.get_routers_by_provider(
            context.session, nconst.ROUTER_PROVIDER_L3AGENT, router_ids)
        # If no l3-agent hosted router, there is no need to schedule.
        if not router_ids:
            return
        return super(L3AgentSchedulerDbMixin, self).auto_schedule_routers(
            context, host, router_ids)

    def schedule_router(self, context, router, candidates=None):
        if (self._get_provider_by_router_id(context, router) ==
                nconst.ROUTER_PROVIDER_L3AGENT):
            return super(L3AgentSchedulerDbMixin, self).schedule_router(
                context, router, candidates=candidates)

    def add_router_to_l3_agent(self, context, id, router_id):
        provider = self._get_provider_by_router_id(context, router_id)
        if provider != nconst.ROUTER_PROVIDER_L3AGENT:
            raise nexc.RouterProviderMismatch(
                router_id=router_id, provider=provider,
                expected_provider=nconst.ROUTER_PROVIDER_L3AGENT)
        return super(L3AgentSchedulerDbMixin, self).add_router_to_l3_agent(
            context, id, router_id)


class L3AgentNotifyAPI(l3_rpc_agent_api.L3AgentNotifyAPI):

    def _notification(self, context, method, router_ids, operation,
                      shuffle_agents, schedule_routers=True):
        """Notify all the agents that are hosting the routers.

        _notification() is called in L3 db plugin for all routers regardless
        the routers are hosted on l3 agents or not. When the routers are
        not hosted on l3 agents, there is no need to notify.
        This method filters routers not hosted by l3 agents.
        """
        router_ids = rdb.get_routers_by_provider(
            context.session, nconst.ROUTER_PROVIDER_L3AGENT, router_ids)
        super(L3AgentNotifyAPI, self)._notification(
            context, method, router_ids, operation, shuffle_agents,
            schedule_routers)


def load_driver(plugin, ofc_manager):

    if (PROVIDER_OPENFLOW in ROUTER_DRIVER_MAP and
            not ofc_manager.driver.router_supported):
        LOG.warning(
            _LW('OFC does not support router with provider=%(provider)s, '
                'so removed it from supported provider '
                '(new router driver map=%(driver_map)s)'),
            {'provider': PROVIDER_OPENFLOW,
             'driver_map': ROUTER_DRIVER_MAP})
        del ROUTER_DRIVER_MAP[PROVIDER_OPENFLOW]

    if cfg.CONF.PROVIDER.default_router_provider not in ROUTER_DRIVER_MAP:
        LOG.error(_LE('default_router_provider %(default)s is supported! '
                      'Please specify one of %(supported)s'),
                  {'default': cfg.CONF.PROVIDER.default_router_provider,
                   'supported': ROUTER_DRIVER_MAP.keys()})
        raise SystemExit(1)

    enabled_providers = (set(cfg.CONF.PROVIDER.router_providers +
                             [cfg.CONF.PROVIDER.default_router_provider]) &
                         set(ROUTER_DRIVER_MAP.keys()))

    for driver in enabled_providers:
        driver_klass = importutils.import_class(ROUTER_DRIVER_MAP[driver])
        ROUTER_DRIVERS[driver] = driver_klass(plugin, ofc_manager)

    LOG.info(_LI('Enabled router drivers: %s'), ROUTER_DRIVERS.keys())

    if not ROUTER_DRIVERS:
        LOG.error(_LE('No router provider is enabled. neutron-server '
                      'terminated! (supported=%(supported)s, '
                      'configured=%(config)s)'),
                  {'supported': ROUTER_DRIVER_MAP.keys(),
                   'config': cfg.CONF.PROVIDER.router_providers})
        raise SystemExit(1)


def get_provider_with_default(provider):
    if not attr.is_attr_set(provider):
        provider = cfg.CONF.PROVIDER.default_router_provider
    elif provider not in ROUTER_DRIVERS:
        raise nexc.ProviderNotFound(provider=provider)
    return provider


def get_driver_by_provider(provider):
    if provider is None:
        provider = cfg.CONF.PROVIDER.default_router_provider
    elif provider not in ROUTER_DRIVERS:
        raise nexc.ProviderNotFound(provider=provider)
    return ROUTER_DRIVERS[provider]



# Copyright 2013 NEC Corporation.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from neutron.api import extensions
from neutron.api.v2 import attributes


LOG = logging.getLogger(__name__)

ROUTER_PROVIDER = 'provider'

ROUTER_PROVIDER_ATTRIBUTE = {
    'routers': {ROUTER_PROVIDER:
                {'allow_post': True,
                 'allow_put': False,
                 'is_visible': True,
                 'default': attributes.ATTR_NOT_SPECIFIED}
                }
}


class Router_provider(extensions.ExtensionDescriptor):
    @classmethod
    def get_name(cls):
        return "Router Provider"

    @classmethod
    def get_alias(cls):
        return "router_provider"

    @classmethod
    def get_description(cls):
        return "Router Provider Support"

    @classmethod
    def get_updated(cls):
        return "2013-08-20T10:00:00-00:00"

    def get_extended_resources(self, version):
        if version == "2.0":
            return ROUTER_PROVIDER_ATTRIBUTE
        else:
            return {}
